"""Constants."""

NOAA_AORC_S3_BASE_URL = "s3://noaa-nws-aorc-v1-1-1km"

AORC_PRECIP_VARIABLE = "APCP_surface"
AORC_X_VAR = "longitude"
AORC_Y_VAR = "latitude"

MM_TO_INCH_CONVERSION_FACTOR = 0.03937007874015748

SHG_WKT = 'PROJCS["USA_Contiguous_Albers_Equal_Area_Conic_USGS_version",GEOGCS["GCS_North_American_1983",DATUM["D_North_American_1983",SPHEROID["GRS_1980",6378137.0,298.257222101]],PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],PROJECTION["Albers"],PARAMETER["False_Easting",0.0],PARAMETER["False_Northing",0.0],PARAMETER["Central_Meridian",-96.0],PARAMETER["Standard_Parallel_1",29.5],PARAMETER["Standard_Parallel_2",45.5],PARAMETER["Latitude_Of_Origin",23.0],UNIT["Meter",1.0]]'
"""CRS definition, based on EPSG:5070"""

KM_TO_M_CONVERSION_FACTOR = 1000
